# -*- coding: utf-8 -*-

"""A getter for HGNC."""

import datetime

from ..utils import Getter, VersionType, get_soup

__all__ = [
    "HGNCGetter",
]


class HGNCGetter(Getter):
    """A getter for HGNC."""

    bioregistry_id = "hgnc"
    name = "HGNC"
    homepage_fmt = "https://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/archive/monthly/json/hgnc_complete_set_{version}.json"

    version_type = VersionType.date

    def get(self) -> str:
        """Get the latest HGNC version number."""
        # TODO scrape https://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/archive/monthly/json instead
        return datetime.date.today().strftime("%Y-%m-01")


if __name__ == "__main__":
    HGNCGetter.print()



# -*- coding: utf-8 -*-

"""A getter for DrugBank."""

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "DrugBankGetter",
]

URL = "https://go.drugbank.com/releases/latest"


class DrugBankGetter(Getter):
    """A getter for DrugBank."""

    bioregistry_id = "drugbank"
    name = "DrugBank"
    homepage_fmt = "https://go.drugbank.com/releases/{version}"
    date_fmt = "%Y-%m-%d"
    version_type = VersionType.semver

    def get(self):
        """Get the latest DrugBank version number."""
        soup = get_soup(URL)
        manifest = soup.find(**{"class": "download-table"}).find("table").find("tbody").find("tr")
        manifest = list(manifest)
        date = manifest[1].text
        version = manifest[2].text
        return dict(date=date, version=version)

    @staticmethod
    def homepage_version_transform(version: str) -> str:
        """Replace dots with dashes for DrugBank homepage format."""
        return version.replace(".", "-")


if __name__ == "__main__":
    DrugBankGetter.print()



# -*- coding: utf-8 -*-

"""A getter for MSigDB."""

from typing import Mapping

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "MSigDBGetter",
]

URL = "http://www.gsea-msigdb.org/gsea/msigdb/index.jsp"


class MSigDBGetter(Getter):
    """A getter for MSigDB."""

    bioregistry_id = "msigdb"
    name = "MSigDB"
    homepage_fmt = "https://data.broadinstitute.org/gsea-msigdb/msigdb/release/{version}"
    date_fmt = "%B %Y"
    version_type = VersionType.month

    def get(self) -> Mapping[str, str]:
        """Get the latest MSigDB version number."""
        soup = get_soup(URL)

        header = soup.find("h1")
        version = header.text.split(" ")[-1].lstrip("v")

        x = soup.find(text="Current Version")
        paragraph = x.parent.find_next_sibling("p")
        date = paragraph.text.strip()[len(f"MSigDB database v{version} updated ") :].split(".")[0]

        return dict(date=date, version=version)


if __name__ == "__main__":
    MSigDBGetter.print()



# -*- coding: utf-8 -*-

"""A getter for miRBase."""

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "MirbaseGetter",
]

PREFIX = "0_THIS_IS_RELEASE_"


class MirbaseGetter(Getter):
    """A getter for miRBase."""

    bioregistry_id = "mirbase"
    name = "miRBase"
    homepage_fmt = "https://www.mirbase.org/ftp/{version}/"
    version_type = VersionType.semver_minor

    def get(self):
        """Get the latest miRBase version number."""
        url = "https://www.mirbase.org/ftp/CURRENT/"
        soup = get_soup(url)
        rows = list(soup.find_all("tr"))
        cells = list(rows[3].find_all("td"))
        text = cells[1].find("a").text
        return text[len(PREFIX) :]


if __name__ == "__main__":
    MirbaseGetter.print()



# -*- coding: utf-8 -*-

"""A getter for Rfam."""

from bioversions.utils import Getter, VersionType, _get_ftp_version

__all__ = [
    "RfamGetter",
]


class RfamGetter(Getter):
    """A getter for Rfam."""

    bioregistry_id = "rfam"
    name = "Rfam"
    homepage_fmt = "ftp://ftp.ebi.ac.uk/pub/databases/Rfam/{version}/"
    version_type = VersionType.semver_minor

    def get(self):
        """Get the latest Rfam version number."""
        return _get_ftp_version("ftp.ebi.ac.uk", "pub/databases/Rfam/")


if __name__ == "__main__":
    RfamGetter.print()



# -*- coding: utf-8 -*-

"""A getter for the `Drug Gene Interaction Database (DGI-DB) <http://www.dgidb.org>`_."""

import os

import bs4
import requests

from bioversions.utils import Getter, VersionType

DOWNLOADS_PAGE = "https://www.dgidb.org/downloads"


class DGIGetter(Getter):
    """A getter for DGI."""

    name = "Drug Gene Interaction Database"
    version_type = VersionType.month
    date_version_fmt = "%Y-%b"

    def get(self):
        """Get the latest DGI version number."""
        res = requests.get(DOWNLOADS_PAGE)
        soup = bs4.BeautifulSoup(res.content, parser="lxml", features="lxml")
        cells = list(soup.select("table#tsv_downloads tbody tr:first-child td:nth-child(2) a"))
        if 1 != len(cells):
            raise ValueError
        cell = cells[0]
        href = cell["href"]
        version = os.path.dirname(os.path.relpath(href, "data/monthly_tsvs"))
        return version


if __name__ == "__main__":
    DGIGetter.print()



# -*- coding: utf-8 -*-

"""A getter for DrugCentral."""

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "DrugCentralGetter",
]

URL = "https://drugcentral.org/download"


class DrugCentralGetter(Getter):
    """A getter for DrugCentral."""

    bioregistry_id = "drugcentral"
    name = "DrugCentral"
    date_version_fmt = "%Y_%m_%d"
    homepage_fmt = href = "https://unmtid-shinyapps.net/download/DrugCentral/{version}"
    version_type = VersionType.date

    def get(self) -> str:
        """Get the latest DrugCentral version number."""
        soup = get_soup(URL)
        manifest = soup.find("h4", **{"class": "txt-sec"}).find("a")
        return manifest.attrs["href"].split("/")[-2]


if __name__ == "__main__":
    DrugCentralGetter.print()



# -*- coding: utf-8 -*-

"""A getter for BioGRID."""

from ..utils import Getter, VersionType, get_soup

__all__ = [
    "BioGRIDGetter",
]

URL = "https://downloads.thebiogrid.org/BioGRID/Latest-Release/"


class BioGRIDGetter(Getter):
    """A getter for BioGRID."""

    bioregistry_id = "biogrid"
    name = "BioGRID"
    homepage_fmt = "https://downloads.thebiogrid.org/BioGRID/Release-Archive/BIOGRID-{version}"
    version_type = VersionType.semver

    def get(self) -> str:
        """Get the latest BioGRID version number."""
        soup = get_soup(URL)
        manifest = soup.find(id="manifestDesc")
        header = manifest.find("h2")
        return header.text[len("BioGRID Release ") :]


if __name__ == "__main__":
    BioGRIDGetter.print()



# -*- coding: utf-8 -*-

"""Utilities and implementation for bioversions."""

import datetime
import enum
import ftplib
import os
from typing import Any, ClassVar, Mapping, Optional, Union

import bioregistry
import pydantic
import pystow
import requests
import requests_ftp
from bs4 import BeautifulSoup
from cachier import cachier

BIOVERSIONS_HOME = pystow.join("bioversions")
HERE = os.path.abspath(os.path.dirname(__file__))
DOCS = os.path.abspath(os.path.join(HERE, os.pardir, os.pardir, "docs"))
IMG = os.path.join(DOCS, "img")

requests_ftp.monkeypatch_session()


class VersionType(enum.Enum):
    """Different types of versions."""

    semver = "SemVer (X.Y.Z)"
    date = "CalVer (YYYY-MM-DD)"
    month = "CalVer (YYYY-MM)"
    year = "CalVer (YYYY)"
    year_minor = "CalVer (YYYY.X)"
    semver_minor = "SemVer (X.Y)"
    sequential = "Sequential (X)"
    daily = "Daily"
    unversioned = "Unversioned"
    other = "Other"
    missing = "Missing"
    #: Saved for the most shameful of data
    garbage = "Garbage"


def norm(s: str) -> str:
    """Normalize a string for dictionary lookup."""
    return s.lower().replace(" ", "").replace("-", "").replace(".", "")


def get_soup(url: str, verify: bool = True) -> BeautifulSoup:
    """Get a beautiful soup parsed version of the given web page."""
    res = requests.get(url, verify=verify)
    soup = BeautifulSoup(res.text, features="html.parser")
    return soup


#: A decorator for functions whose return values
#: should be cached and refreshed once per day
refresh_daily = cachier(
    stale_after=datetime.timedelta(days=1),
    backend="memory",
    cache_dir=BIOVERSIONS_HOME,
)


class MetaGetter(type):
    """A metatype to expose two class properties."""

    _cache = None

    date_fmt: Optional[str]
    date_version_fmt: Optional[str]
    homepage_fmt: Optional[str]

    @property
    def _cache_prop(cls):
        if cls._cache is None:
            cls._cache = cls().get()
        return cls._cache

    @property
    def version(cls) -> str:
        """Get the version of the getter based on the inheriting class's implementation."""
        if isinstance(cls._cache_prop, str):
            return cls._cache_prop
        elif isinstance(cls._cache_prop, dict):
            return cls._cache_prop["version"]
        elif isinstance(cls._cache_prop, datetime.datetime):
            return cls._cache_prop.strftime("%Y-%m-%d")
        else:
            raise TypeError(f"_cache_prop was a {type(cls._cache_prop)}")

    @property
    def date(cls) -> Optional[datetime.date]:
        """Get the date if it's set."""
        vp = cls.version_date_parsed
        if vp is not None:
            return vp
        if not isinstance(cls._cache_prop, dict):
            return None
        date = cls._cache_prop["date"]
        if isinstance(date, datetime.datetime):
            return date.date()
        if not cls.date_fmt:
            raise TypeError(
                f"Need to set {cls.__name__} class variable `date_fmt` to parse date {date}"
            )
        try:
            return datetime.datetime.strptime(date, cls.date_fmt).date()
        except ValueError:
            raise ValueError(f"Issue in {cls.__name__} with date {date} and fmt {cls.date_fmt}")

    @property
    def version_date_parsed(cls) -> Optional[datetime.date]:
        """Get the date as a parsed class there's a format string."""
        if cls.date_version_fmt is None:
            return None
        try:
            return datetime.datetime.strptime(cls.version, cls.date_version_fmt).date()
        except ValueError:
            raise ValueError(
                f"Issue parsing {cls.__name__} version {cls.version} with fmt {cls.date_version_fmt}"
            )

    @property
    def homepage(cls) -> Optional[str]:
        """Get the homepage's URL if a format string was specified."""
        if cls.homepage_fmt is None:
            return None

        version = cls.homepage_version_transform(cls.version)
        return cls.homepage_fmt.format(version=version)

    @staticmethod
    def homepage_version_transform(version: str) -> str:
        """Transform the version for formatting into the homepage."""
        return version


class Bioversion(pydantic.BaseModel):
    """A dataclass for information about a database and version."""

    #: The database name
    name: str
    #: The database current version
    version: str
    #: The class that retrieved the version
    classname: str
    #: The version type
    vtype: VersionType
    #: The date of the current release
    date: Optional[datetime.date]
    #: The URL for the homepage of the specific version of the database
    homepage: Optional[str]
    #: The database prefix
    bioregistry_id: Optional[str]


class Getter(metaclass=MetaGetter):
    """A class for holding the name of a database and implementation of the version getter."""

    #: The name of the database. Specify this in the inheriting class!.
    name: ClassVar[str]
    #: The type of version string. Required!
    version_type: ClassVar[VersionType]

    #: The URL with `{version}` to format in the version. Specify this in the inheriting class.
    homepage_fmt: ClassVar[Optional[str]] = None

    date_fmt: ClassVar[Optional[str]] = None

    date_version_fmt: ClassVar[Optional[str]] = None

    bioregistry_id: ClassVar[Optional[str]] = None

    # The following are automatically calculated based on the metaclass
    version: ClassVar[str]
    date: ClassVar[str]
    homepage: ClassVar[str]

    def get(self) -> Union[str, Mapping[str, str], datetime.datetime]:
        """Get the latest of this database."""
        raise NotImplementedError

    @classmethod
    def print(cls, sep: str = "\t", file=None):
        """Print the latest version of this database."""
        x = [cls.bioregistry_id, cls.name, cls.version]
        if cls.date:
            x.append(f"({cls.date})")
        if cls.homepage:
            x.append(cls.homepage)
        print(*x, sep=sep, file=file)

    @classmethod
    def resolve(cls) -> Bioversion:
        """Get a Bioversion data container with the data for this database."""
        return Bioversion(
            name=cls.name,
            version=cls.version,
            classname=cls.__name__,
            vtype=cls.version_type,
            homepage=cls.homepage,
            date=cls.date,
            bioregistry_id=cls.bioregistry_id,
        )

    @classmethod
    def to_dict(cls) -> Mapping[str, Any]:
        """Get a dict with the data for this database."""
        return cls.resolve().to_dict()


class DailyGetter(Getter):
    """A base getter for daily updated resources."""

    version_type = VersionType.daily

    def get(self) -> Union[str, Mapping[str, str]]:
        """Return a constant "daily" string."""
        return "daily"


class UnversionedGetter(Getter):
    """A base getter for unversioned resources."""

    version_type = VersionType.unversioned

    #: Has this database been apparently abandoned (true) or is it still updated (false)
    abandoned: ClassVar[bool]

    def get(self) -> Union[str, Mapping[str, str]]:
        """Return a constant unversioned string."""
        return "unversioned"


def get_obo_version(url: str) -> str:
    """Get the data version from an OBO file."""
    with requests.get(url, stream=True) as res:
        for line in res.iter_lines():
            line = line.decode("utf-8")
            if line.startswith("data-version:"):
                version = line[len("data-version:") :].strip()
                return version
    raise ValueError(f"No data-version line contained in {url}")


class OboGetter(Getter):
    """An implementation for getting OBO versions."""

    strip_key_prefix: ClassVar[bool] = False
    strip_version_prefix: ClassVar[bool] = False
    strip_file_suffix: ClassVar[bool] = False

    @property
    def key(self) -> str:
        """Get the OBO Foundry key."""
        rv = bioregistry.get_obofoundry_prefix(self.bioregistry_id)
        if rv is None:
            raise ValueError
        return rv

    def get(self) -> str:
        """Get the OBO version."""
        url = f"http://purl.obolibrary.org/obo/{self.key}.obo"
        return self.process(get_obo_version(url))

    def process(self, version: str) -> str:
        """Post-process the version string."""
        if self.strip_key_prefix:
            version = version[len(f"{self.key}/") :]
        if self.strip_version_prefix:
            version = version[len("releases/") :]
        if self.strip_file_suffix:
            version = version[: -(len(self.key) + 5)]
        return version


def _get_ftp_version(host: str, directory: str) -> str:
    with ftplib.FTP(host) as ftp:
        ftp.login()
        ftp.cwd(directory)
        names = sorted(
            [
                tuple(int(part) for part in name.split("."))
                for name in ftp.nlst()
                if _is_version(name)
            ]
        )
    return ".".join(map(str, names[-1]))


def _get_ftp_date_version(host: str, directory: str) -> str:
    with ftplib.FTP(host) as ftp:
        ftp.login()
        ftp.cwd(directory)
        names = sorted([name for name in ftp.nlst() if _is_iso_8601(name)])
    return names[-1]


def _is_iso_8601(s: str) -> bool:
    x = s.split("-")
    return len(x) == 3 and x[0].isnumeric() and x[1].isnumeric() and x[2].isnumeric()


def _is_version(s: str) -> bool:
    x = s.split(".")
    return len(x) == 2 and x[0].isnumeric() and x[1].isnumeric()


def _is_semantic_version(s: str) -> bool:
    x = s.split(".")
    return len(x) == 3 and x[0].isnumeric() and x[1].isnumeric() and x[2].isnumeric()



# -*- coding: utf-8 -*-

"""Getters for OBO ontologies."""

from typing import Iterable, Type

from bioversions.utils import Getter, OboGetter, VersionType

__all__ = [
    "iter_obo_getters",
    "ChebiGetter",
    "GoGetter",
    "DoidGetter",
    "PrGetter",
]


class ChebiGetter(OboGetter):
    """A getter for ChEBI."""

    name = "ChEBI"
    bioregistry_id = "chebi"
    version_type = VersionType.sequential
    homepage_fmt = "ftp://ftp.ebi.ac.uk/pub/databases/chebi/archive/rel{version}/"


class GoGetter(OboGetter):
    """A getter for the Gene Ontology (GO)."""

    name = "Gene Ontology"
    bioregistry_id = "go"
    version_type = VersionType.date
    strip_version_prefix = True
    date_version_fmt = "%Y-%m-%d"
    homepage_fmt = "http://archive.geneontology.org/full/{version}/"


class DoidGetter(OboGetter):
    """A getter for the Disease Ontology (DO)."""

    name = "Disease Ontology"
    homepage_fmt = "https://github.com/DiseaseOntology/HumanDiseaseOntology/tree/main/src/ontology/releases/{version}"
    bioregistry_id = "doid"
    version_type = VersionType.date
    strip_version_prefix = True
    strip_file_suffix = True
    date_version_fmt = "%Y-%m-%d"


class PrGetter(OboGetter):
    """A getter for the Protein Ontology (PR)."""

    name = "Protein Ontology"
    bioregistry_id = "pr"
    version_type = VersionType.semver_minor
    homepage_fmt = "https://proconsortium.org/download/release_{version}/"


def iter_obo_getters() -> Iterable[Type[Getter]]:
    """Iterate over OBO getters."""
    yield from OboGetter.__subclasses__()


def _main():
    for getter in iter_obo_getters():
        getter.print()


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""A getter for MeSH."""

import ftplib

from bioversions.utils import Getter, VersionType

__all__ = [
    "MeshGetter",
]


class MeshGetter(Getter):
    """A getter for MeSH."""

    bioregistry_id = "mesh"
    name = "MeSH"
    homepage_fmt = "ftp://nlmpubs.nlm.nih.gov/online/mesh/{version}"
    version_type = VersionType.year

    def get(self):
        """Get the latest MeSH version number."""
        with ftplib.FTP("nlmpubs.nlm.nih.gov") as ftp:
            ftp.login()
            ftp.cwd("/online/mesh/MESH_FILES/xmlmesh/")
            for name, _ in ftp.mlsd():
                if name.startswith("desc") and name.endswith(".gz"):
                    return name[len("desc") : -len(".gz")]
        raise ValueError


if __name__ == "__main__":
    MeshGetter.print()
